# -*- coding: utf-8 -*-
'''
Python interface to the awesome mpv media player

Created on 2017-02-19
@author: jaseg
forked from: https://github.com/jaseg/python-mpv
'''

from ctypes import *
import ctypes.util
import threading
import os
import sys
from warnings import warn
from functools import partial
import collections
import re
import traceback

# vim: ts=4 sw=4 et

if os.name == 'nt':
    backend = CDLL('mpv-1.dll')
    fs_enc = 'utf-8'
else:
    import locale
    lc, enc = locale.getlocale(locale.LC_NUMERIC)
    # libmpv requires LC_NUMERIC to be set to "C". Since messing with global variables everyone else relies upon is
    # still better than segfaulting, we are setting LC_NUMERIC to "C".
    locale.setlocale(locale.LC_NUMERIC, 'C')

    sofile = ctypes.util.find_library('mpv')
    if sofile is None:
        raise OSError("Cannot find libmpv in the usual places. Depending on your distro, you may try installing an "
                "mpv-devel or mpv-libs package. If you have libmpv around but this script can't find it, maybe consult "
                "the documentation for ctypes.util.find_library which this script uses to look up the library "
                "filename.")
    backend = CDLL(sofile)
    fs_enc = sys.getfilesystemencoding()


class MpvHandle(c_void_p):
    pass

class MpvOpenGLCbContext(c_void_p):
    pass


class PropertyUnavailableError(AttributeError):
    pass

class ErrorCode(object):
    """ For documentation on these, see mpv's libmpv/client.h """
    SUCCESS                 = 0
    EVENT_QUEUE_FULL        = -1
    NOMEM                   = -2
    UNINITIALIZED           = -3
    INVALID_PARAMETER       = -4
    OPTION_NOT_FOUND        = -5
    OPTION_FORMAT           = -6
    OPTION_ERROR            = -7
    PROPERTY_NOT_FOUND      = -8
    PROPERTY_FORMAT         = -9
    PROPERTY_UNAVAILABLE    = -10
    PROPERTY_ERROR          = -11
    COMMAND                 = -12

    EXCEPTION_DICT = {
             0:     None,
            -1:     lambda *a: MemoryError('mpv event queue full', *a),
            -2:     lambda *a: MemoryError('mpv cannot allocate memory', *a),
            -3:     lambda *a: ValueError('Uninitialized mpv handle used', *a),
            -4:     lambda *a: ValueError('Invalid value for mpv parameter', *a),
            -5:     lambda *a: AttributeError('mpv option does not exist', *a),
            -6:     lambda *a: TypeError('Tried to set mpv option using wrong format', *a),
            -7:     lambda *a: ValueError('Invalid value for mpv option', *a),
            -8:     lambda *a: AttributeError('mpv property does not exist', *a),
            # Currently (mpv 0.18.1) there is a bug causing a PROPERTY_FORMAT error to be returned instead of
            # INVALID_PARAMETER when setting a property-mapped option to an invalid value.
            -9:     lambda *a: TypeError('Tried to get/set mpv property using wrong format, or passed invalid value', *a),
            -10:    lambda *a: PropertyUnavailableError('mpv property is not available', *a),
            -11:    lambda *a: RuntimeError('Generic error getting or setting mpv property', *a),
            -12:    lambda *a: SystemError('Error running mpv command', *a) }

    @staticmethod
    def default_error_handler(ec, *args):
        return ValueError(_mpv_error_string(ec).decode('utf-8'), ec, *args)

    @classmethod
    def raise_for_ec(kls, ec, func, *args):
        ec = 0 if ec > 0 else ec
        ex = kls.EXCEPTION_DICT.get(ec , kls.default_error_handler)
        if ex:
            raise ex(ec, *args)


class MpvFormat(c_int):
    NONE        = 0
    STRING      = 1
    OSD_STRING  = 2
    FLAG        = 3
    INT64       = 4
    DOUBLE      = 5
    NODE        = 6
    NODE_ARRAY  = 7
    NODE_MAP    = 8
    BYTE_ARRAY  = 9

    def __eq__(self, other):
        return self is other or self.value == other or self.value == int(other)

    def __repr__(self):
        return ['NONE', 'STRING', 'OSD_STRING', 'FLAG', 'INT64', 'DOUBLE', 'NODE', 'NODE_ARRAY', 'NODE_MAP',
                'BYTE_ARRAY'][self.value]



class MpvEventID(c_int):
    NONE                    = 0
    SHUTDOWN                = 1
    LOG_MESSAGE             = 2
    GET_PROPERTY_REPLY      = 3
    SET_PROPERTY_REPLY      = 4
    COMMAND_REPLY           = 5
    START_FILE              = 6
    END_FILE                = 7
    FILE_LOADED             = 8
    TRACKS_CHANGED          = 9
    TRACK_SWITCHED          = 10
    IDLE                    = 11
    PAUSE                   = 12
    UNPAUSE                 = 13
    TICK                    = 14
    SCRIPT_INPUT_DISPATCH   = 15
    CLIENT_MESSAGE          = 16
    VIDEO_RECONFIG          = 17
    AUDIO_RECONFIG          = 18
    METADATA_UPDATE         = 19
    SEEK                    = 20
    PLAYBACK_RESTART        = 21
    PROPERTY_CHANGE         = 22
    CHAPTER_CHANGE          = 23

    ANY = ( SHUTDOWN, LOG_MESSAGE, GET_PROPERTY_REPLY, SET_PROPERTY_REPLY, COMMAND_REPLY, START_FILE, END_FILE,
            FILE_LOADED, TRACKS_CHANGED, TRACK_SWITCHED, IDLE, PAUSE, UNPAUSE, TICK, SCRIPT_INPUT_DISPATCH,
            CLIENT_MESSAGE, VIDEO_RECONFIG, AUDIO_RECONFIG, METADATA_UPDATE, SEEK, PLAYBACK_RESTART, PROPERTY_CHANGE,
            CHAPTER_CHANGE )

    def __repr__(self):
        return ['NONE', 'SHUTDOWN', 'LOG_MESSAGE', 'GET_PROPERTY_REPLY', 'SET_PROPERTY_REPLY', 'COMMAND_REPLY',
                'START_FILE', 'END_FILE', 'FILE_LOADED', 'TRACKS_CHANGED', 'TRACK_SWITCHED', 'IDLE', 'PAUSE', 'UNPAUSE',
                'TICK', 'SCRIPT_INPUT_DISPATCH', 'CLIENT_MESSAGE', 'VIDEO_RECONFIG', 'AUDIO_RECONFIG',
                'METADATA_UPDATE', 'SEEK', 'PLAYBACK_RESTART', 'PROPERTY_CHANGE', 'CHAPTER_CHANGE'][self.value]


class MpvNodeList(Structure):
    def array_value(self, decode_str=False):
        return [ self.values[i].node_value(decode_str) for i in range(self.num) ]

    def dict_value(self, decode_str=False):
        return { self.keys[i].decode('utf-8'): self.values[i].node_value(decode_str) for i in range(self.num) }

class MpvNode(Structure):
    _fields_ = [('val', c_longlong),
                ('format', MpvFormat)]

    def node_value(self, decode_str=False):
        return MpvNode.node_cast_value(byref(c_void_p(self.val)), self.format.value, decode_str)

    @staticmethod
    def node_cast_value(v, fmt, decode_str=False):
        dwrap = lambda s: s.decode('utf-8') if decode_str else s
        return {
            MpvFormat.NONE:         lambda v: None,
            MpvFormat.STRING:       lambda v: dwrap(cast(v, POINTER(c_char_p)).contents.value),
            MpvFormat.OSD_STRING:   lambda v: cast(v, POINTER(c_char_p)).contents.value.decode('utf-8'),
            MpvFormat.FLAG:         lambda v: bool(cast(v, POINTER(c_int)).contents.value),
            MpvFormat.INT64:        lambda v: cast(v, POINTER(c_longlong)).contents.value,
            MpvFormat.DOUBLE:       lambda v: cast(v, POINTER(c_double)).contents.value,
            MpvFormat.NODE:         lambda v: cast(v, POINTER(MpvNode)).contents.node_value(decode_str),
            MpvFormat.NODE_ARRAY:   lambda v: cast(v, POINTER(POINTER(MpvNodeList))).contents.contents.array_value(decode_str),
            MpvFormat.NODE_MAP:     lambda v: cast(v, POINTER(POINTER(MpvNodeList))).contents.contents.dict_value(decode_str),
            MpvFormat.BYTE_ARRAY:   lambda v: cast(v, POINTER(c_char_p)).contents.value,
            }[fmt](v)

MpvNodeList._fields_ = [('num', c_int),
                        ('values', POINTER(MpvNode)),
                        ('keys', POINTER(c_char_p))]

class MpvSubApi(c_int):
    MPV_SUB_API_OPENGL_CB   = 1

class MpvEvent(Structure):
    _fields_ = [('event_id', MpvEventID),
                ('error', c_int),
                ('reply_userdata', c_ulonglong),
                ('data', c_void_p)]

    def as_dict(self):
        dtype = {MpvEventID.END_FILE:               MpvEventEndFile,
                MpvEventID.PROPERTY_CHANGE:         MpvEventProperty,
                MpvEventID.GET_PROPERTY_REPLY:      MpvEventProperty,
                MpvEventID.LOG_MESSAGE:             MpvEventLogMessage,
                MpvEventID.SCRIPT_INPUT_DISPATCH:   MpvEventScriptInputDispatch,
                MpvEventID.CLIENT_MESSAGE:          MpvEventClientMessage
            }.get(self.event_id.value, None)
        return {'event_id': self.event_id.value,
                'error': self.error,
                'reply_userdata': self.reply_userdata,
                'event': cast(self.data, POINTER(dtype)).contents.as_dict() if dtype else None}

class MpvEventProperty(Structure):
    _fields_ = [('name', c_char_p),
                ('format', MpvFormat),
                ('data', c_void_p)]
    def as_dict(self):
        if self.format.value == MpvFormat.STRING:
            proptype, _access = ALL_PROPERTIES.get(self.name, (str, None))
            return {'name': self.name.decode('utf-8'),
                    'format': self.format,
                    'data': self.data,
                    'value': proptype(cast(self.data, POINTER(c_char_p)).contents.value.decode('utf-8'))}
        else:
            return {'name': self.name.decode('utf-8'),
                    'format': self.format,
                    'data': self.data}

class MpvEventLogMessage(Structure):
    _fields_ = [('prefix', c_char_p),
                ('level', c_char_p),
                ('text', c_char_p)]

    def as_dict(self):
        return { 'prefix': self.prefix.decode('utf-8'),
                 'level':  self.level.decode('utf-8'),
                 'text':   self.text.decode('utf-8').rstrip() }

class MpvEventEndFile(c_int):
    EOF_OR_INIT_FAILURE = 0
    RESTARTED           = 1
    ABORTED             = 2
    QUIT                = 3

    def as_dict(self):
        return {'reason': self.value}

class MpvEventScriptInputDispatch(Structure):
    _fields_ = [('arg0', c_int),
                ('type', c_char_p)]

    def as_dict(self):
        pass # TODO

class MpvEventClientMessage(Structure):
    _fields_ = [('num_args', c_int),
                ('args', POINTER(c_char_p))]

    def as_dict(self):
        return { 'args': [ self.args[i].decode('utf-8') for i in range(self.num_args) ] }

WakeupCallback = CFUNCTYPE(None, c_void_p)

OpenGlCbUpdateFn = CFUNCTYPE(None, c_void_p)
OpenGlCbGetProcAddrFn = CFUNCTYPE(None, c_void_p, c_char_p)

def _handle_func(name, args, restype, errcheck, ctx=MpvHandle):
    func = getattr(backend, name)
    func.argtypes = [ctx] + args if ctx else args
    if restype is not None:
        func.restype = restype
    if errcheck is not None:
        func.errcheck = errcheck
    globals()['_'+name] = func

def bytes_free_errcheck(res, func, *args):
    notnull_errcheck(res, func, *args)
    rv = cast(res, c_void_p).value
    _mpv_free(res)
    return rv

def notnull_errcheck(res, func, *args):
    if res is None:
        raise RuntimeError('Underspecified error in MPV when calling {} with args {!r}: NULL pointer returned.'\
                'Please consult your local debugger.'.format(func.__name__, args))
    return res

ec_errcheck = ErrorCode.raise_for_ec

def _handle_gl_func(name, args=[], restype=None):
    _handle_func(name, args, restype, errcheck=None, ctx=MpvOpenGLCbContext)

backend.mpv_client_api_version.restype = c_ulong
def _mpv_client_api_version():
    ver = backend.mpv_client_api_version()
    return ver>>16, ver&0xFFFF

backend.mpv_free.argtypes = [c_void_p]
_mpv_free = backend.mpv_free

backend.mpv_free_node_contents.argtypes = [c_void_p]
_mpv_free_node_contents = backend.mpv_free_node_contents

backend.mpv_create.restype = MpvHandle
_mpv_create = backend.mpv_create

_handle_func('mpv_create_client',           [c_char_p],                                 MpvHandle, notnull_errcheck)
_handle_func('mpv_client_name',             [],                                         c_char_p, errcheck=None)
_handle_func('mpv_initialize',              [],                                         c_int, ec_errcheck)
_handle_func('mpv_detach_destroy',          [],                                         None, errcheck=None)
_handle_func('mpv_terminate_destroy',       [],                                         None, errcheck=None)
_handle_func('mpv_load_config_file',        [c_char_p],                                 c_int, ec_errcheck)
_handle_func('mpv_suspend',                 [],                                         None, errcheck=None)
_handle_func('mpv_resume',                  [],                                         None, errcheck=None)
_handle_func('mpv_get_time_us',             [],                                         c_ulonglong, errcheck=None)

_handle_func('mpv_set_option',              [c_char_p, MpvFormat, c_void_p],            c_int, ec_errcheck)
_handle_func('mpv_set_option_string',       [c_char_p, c_char_p],                       c_int, ec_errcheck)

_handle_func('mpv_command',                 [POINTER(c_char_p)],                        c_int, ec_errcheck)
_handle_func('mpv_command_string',          [c_char_p, c_char_p],                       c_int, ec_errcheck)
_handle_func('mpv_command_async',           [c_ulonglong, POINTER(c_char_p)],           c_int, ec_errcheck)

_handle_func('mpv_set_property',            [c_char_p, MpvFormat, c_void_p],            c_int, ec_errcheck)
_handle_func('mpv_set_property_string',     [c_char_p, c_char_p],                       c_int, ec_errcheck)
_handle_func('mpv_set_property_async',      [c_ulonglong, c_char_p, MpvFormat,c_void_p],c_int, ec_errcheck)
_handle_func('mpv_get_property',            [c_char_p, MpvFormat, c_void_p],            c_int, ec_errcheck)
_handle_func('mpv_get_property_string',     [c_char_p],                                 c_void_p, bytes_free_errcheck)
_handle_func('mpv_get_property_osd_string', [c_char_p],                                 c_void_p, bytes_free_errcheck)
_handle_func('mpv_get_property_async',      [c_ulonglong, c_char_p, MpvFormat],         c_int, ec_errcheck)
_handle_func('mpv_observe_property',        [c_ulonglong, c_char_p, MpvFormat],         c_int, ec_errcheck)
_handle_func('mpv_unobserve_property',      [c_ulonglong],                              c_int, ec_errcheck)

_handle_func('mpv_event_name',              [c_int],                                    c_char_p, errcheck=None, ctx=None)
_handle_func('mpv_error_string',            [c_int],                                    c_char_p, errcheck=None, ctx=None)

_handle_func('mpv_request_event',           [MpvEventID, c_int],                        c_int, ec_errcheck)
_handle_func('mpv_request_log_messages',    [c_char_p],                                 c_int, ec_errcheck)
_handle_func('mpv_wait_event',              [c_double],                                 POINTER(MpvEvent), errcheck=None)
_handle_func('mpv_wakeup',                  [],                                         None, errcheck=None)
_handle_func('mpv_set_wakeup_callback',     [WakeupCallback, c_void_p],                 None, errcheck=None)
_handle_func('mpv_get_wakeup_pipe',         [],                                         c_int, errcheck=None)

_handle_func('mpv_get_sub_api',             [MpvSubApi],                                c_void_p, notnull_errcheck)

_handle_gl_func('mpv_opengl_cb_set_update_callback',    [OpenGlCbUpdateFn, c_void_p])
_handle_gl_func('mpv_opengl_cb_init_gl',                [c_char_p, OpenGlCbGetProcAddrFn, c_void_p],    c_int)
_handle_gl_func('mpv_opengl_cb_draw',                   [c_int, c_int, c_int],                          c_int)
_handle_gl_func('mpv_opengl_cb_render',                 [c_int, c_int],                                 c_int)
_handle_gl_func('mpv_opengl_cb_report_flip',            [c_ulonglong],                                  c_int)
_handle_gl_func('mpv_opengl_cb_uninit_gl',              [],                                             c_int)


def _ensure_encoding(possibly_bytes):
    return possibly_bytes.decode('utf-8') if type(possibly_bytes) is bytes else possibly_bytes


def _event_generator(handle):
    while True:
        event = _mpv_wait_event(handle, -1).contents
        if event.event_id.value == MpvEventID.NONE:
            raise StopIteration()
        yield event

def load_lua():
    """ Use this function if you intend to use mpv's built-in lua interpreter. This is e.g. needed for playback of
    youtube urls. """
    CDLL('liblua.so', mode=RTLD_GLOBAL)


def _event_loop(event_handle, playback_cond, event_callbacks, message_handlers, property_handlers, log_handler):
    for event in _event_generator(event_handle):
        try:
            devent = event.as_dict() # copy data from ctypes
            eid = devent['event_id']
            for callback in event_callbacks:
                callback(devent)
            if eid in (MpvEventID.SHUTDOWN, MpvEventID.END_FILE):
                with playback_cond:
                    playback_cond.notify_all()
            if eid == MpvEventID.PROPERTY_CHANGE:
                pc = devent['event']
                name = pc['name']

                if 'value' in pc:
                    proptype, _access = ALL_PROPERTIES[name]
                    if proptype is bytes:
                        args = (pc['value'],)
                    else:
                        args = (proptype(_ensure_encoding(pc['value'])),)
                elif pc['format'] == MpvFormat.NONE:
                    args = (None,)
                else:
                    args = (pc['data'], pc['format'])

                for handler in property_handlers[name]:
                    handler(*args)
            if eid == MpvEventID.LOG_MESSAGE and log_handler is not None:
                ev = devent['event']
                log_handler(ev['level'], ev['prefix'], ev['text'])
            if eid == MpvEventID.CLIENT_MESSAGE:
                # {'event': {'args': ['key-binding', 'foo', 'u-', 'g']}, 'reply_userdata': 0, 'error': 0, 'event_id': 16}
                target, *args = devent['event']['args']
                if target in message_handlers:
                    message_handlers[target](*args)
            if eid == MpvEventID.SHUTDOWN:
                _mpv_detach_destroy(event_handle)
                return
        except Exception as e:
            traceback.print_exc()

class MPV(object):
    """ See man mpv(1) for the details of the implemented commands. """
    def __init__(self, *extra_mpv_flags, log_handler=None, start_event_thread=True, **extra_mpv_opts):
        """ Create an MPV instance.

        Extra arguments and extra keyword arguments will be passed to mpv as options. """

        self._event_thread = None
        self.handle = _mpv_create()

        _mpv_set_option_string(self.handle, b'audio-display', b'no')
        istr = lambda o: ('yes' if o else 'no') if type(o) is bool else str(o)
        try:
            for flag in extra_mpv_flags:
                _mpv_set_option_string(self.handle, flag.encode('utf-8'), b'')
            for k,v in extra_mpv_opts.items():
                _mpv_set_option_string(self.handle, k.replace('_', '-').encode('utf-8'), istr(v).encode('utf-8'))
        finally:
            _mpv_initialize(self.handle)

        self._event_callbacks = []
        self._property_handlers = collections.defaultdict(lambda: [])
        self._message_handlers = {}
        self._key_binding_handlers = {}
        self._playback_cond = threading.Condition()
        self._event_handle = _mpv_create_client(self.handle, b'py_event_handler')
        self._loop = partial(_event_loop, self._event_handle, self._playback_cond, self._event_callbacks,
                self._message_handlers, self._property_handlers, log_handler)
        if start_event_thread:
            self._event_thread = threading.Thread(target=self._loop, name='MPVEventHandlerThread')
            self._event_thread.setDaemon(True)
            self._event_thread.start()
        else:
            self._event_thread = None

        if log_handler is not None:
            self.set_loglevel('terminal-default')

    def wait_for_playback(self):
        """ Waits until playback of the current title is paused or done """
        with self._playback_cond:
            self._playback_cond.wait()

    def wait_for_property(self, name, cond=lambda val: val, level_sensitive=True):
        sema = threading.Semaphore(value=0)
        def observer(val):
            if cond(val):
                sema.release()
        self.observe_property(name, observer)
        if not level_sensitive or not cond(getattr(self, name.replace('-', '_'))):
            sema.acquire()
        self.unobserve_property(name, observer)

    def __del__(self):
        if self.handle:
            self.terminate()

    def terminate(self):
        self.handle, handle = None, self.handle
        if threading.current_thread() is self._event_thread:
            # Handle special case to allow event handle to be detached.
            # This is necessary since otherwise the event thread would deadlock itself.
            grim_reaper = threading.Thread(target=lambda: _mpv_terminate_destroy(handle))
            grim_reaper.start()
        else:
            _mpv_terminate_destroy(handle)
            if self._event_thread:
                self._event_thread.join()

    def set_loglevel(self, level):
        _mpv_request_log_messages(self._event_handle, level.encode('utf-8'))

    def command(self, name, *args):
        """ Execute a raw command """
        args = [name.encode('utf-8')] + [ (arg if type(arg) is bytes else str(arg).encode('utf-8'))
                for arg in args if arg is not None ] + [None]
        _mpv_command(self.handle, (c_char_p*len(args))(*args))

    def seek(self, amount, reference="relative", precision="default-precise"):
        self.command('seek', amount, reference, precision)

    def revert_seek(self):
        self.command('revert_seek');

    def frame_step(self):
        self.command('frame_step')

    def frame_back_step(self):
        self.command('frame_back_step')

    def _add_property(self, name, value=None):
        self.command('add_property', name, value)

    def _cycle_property(self, name, direction='up'):
        self.command('cycle_property', name, direction)

    def _multiply_property(self, name, factor):
        self.command('multiply_property', name, factor)

    def screenshot(self, includes='subtitles', mode='single'):
        self.command('screenshot', includes, mode)

    def screenshot_to_file(self, filename, includes='subtitles'):
        self.command('screenshot_to_file', filename.encode(fs_enc), includes)

    def playlist_next(self, mode='weak'):
        self.command('playlist_next', mode)

    def playlist_prev(self, mode='weak'):
        self.command('playlist_prev', mode)

    @staticmethod
    def _encode_options(options):
        return ','.join('{}={}'.format(str(key), str(val)) for key, val in options.items())

    def loadfile(self, filename, mode='replace', **options):
        self.command('loadfile', filename.encode(fs_enc), mode, MPV._encode_options(options))

    def loadlist(self, playlist, mode='replace'):
        self.command('loadlist', playlist.encode(fs_enc), mode)

    def playlist_clear(self):
        self.command('playlist_clear')

    def playlist_remove(self, index='current'):
        self.command('playlist_remove', index)

    def playlist_move(self, index1, index2):
        self.command('playlist_move', index1, index2)

    def run(self, command, *args):
        self.command('run', command, *args)

    def quit(self, code=None):
        self.command('quit', code)

    def quit_watch_later(self, code=None):
        self.command('quit_watch_later', code)

    def sub_add(self, filename):
        self.command('sub_add', filename.encode(fs_enc))

    def sub_remove(self, sub_id=None):
        self.command('sub_remove', sub_id)

    def sub_reload(self, sub_id=None):
        self.command('sub_reload', sub_id)

    def sub_step(self, skip):
        self.command('sub_step', skip)

    def sub_seek(self, skip):
        self.command('sub_seek', skip)

    def toggle_osd(self):
        self.command('osd')

    def show_text(self, string, duration='-', level=None):
        self.command('show_text', string, duration, level)

    def show_progress(self):
        self.command('show_progress')

    def discnav(self, command):
        self.command('discnav', command)

    def write_watch_later_config(self):
        self.command('write_watch_later_config')

    def overlay_add(self, overlay_id, x, y, file_or_fd, offset, fmt, w, h, stride):
        self.command('overlay_add', overlay_id, x, y, file_or_fd, offset, fmt, w, h, stride)

    def overlay_remove(self, overlay_id):
        self.command('overlay_remove', overlay_id)

    def script_message(self, *args):
        self.command('script_message', *args)

    def script_message_to(self, target, *args):
        self.command('script_message_to', target, *args)

    def observe_property(self, name, handler):
        self._property_handlers[name].append(handler)
        _mpv_observe_property(self._event_handle, hash(name)&0xffffffffffffffff, name.encode('utf-8'), MpvFormat.STRING)

    def unobserve_property(self, name, handler):
        handlers = self._property_handlers[name]
        handlers.remove(handler)
        if not handlers:
            _mpv_unobserve_property(self._event_handle, hash(name)&0xffffffffffffffff)

    def register_message_handler(self, target, handler):
        self._message_handlers[target] = handler

    def unregister_message_handler(self, target):
        del self._message_handlers[target]

    def register_event_callback(self, callback):
        self._event_callbacks.append(callback)

    def unregister_event_callback(self, callback):
        self._event_callbacks.remove(callback)

    @staticmethod
    def _binding_name(callback_or_cmd):
        return 'py_kb_{:016x}'.format(hash(callback_or_cmd)&0xffffffffffffffff)

    def register_key_binding(self, keydef, callback_or_cmd, mode='force'):
        """ BIG FAT WARNING: mpv's key binding mechanism is pretty powerful. This means, you essentially get arbitrary
        code exectution through key bindings. This interface makes some limited effort to sanitize the keydef given in
        the first parameter, but YOU SHOULD NOT RELY ON THIS IN FOR SECURITY. If your input comes from config files,
        this is completely fine--but, if you are about to pass untrusted input into this parameter, better double-check
        whether this is secure in your case. """
        if not re.match(r'(Shift+)?(Ctrl+)?(Alt+)?(Meta+)?(.|\w+)', keydef):
            raise ValueError('Invalid keydef. Expected format: [Shift+][Ctrl+][Alt+][Meta+]<key>\n'
                    '<key> is either the literal character the key produces (ASCII or Unicode character), or a '
                    'symbolic name (as printed by --input-keylist')
        binding_name = MPV._binding_name(keydef)
        if callable(callback_or_cmd):
            self._key_binding_handlers[binding_name] = callback_or_cmd
            self.register_message_handler('key-binding', self._handle_key_binding_message)
            self.command('define-section',
                    binding_name, '{} script-binding py_event_handler/{}'.format(keydef, binding_name), mode)
        elif isinstance(callback_or_cmd, str):
            self.command('define-section', binding_name, '{} {}'.format(keydef, callback_or_cmd), mode)
        else:
            raise TypeError('register_key_binding expects either an str with an mpv command or a python callable.')
        self.command('enable-section', binding_name)

    def _handle_key_binding_message(self, binding_name, key_state, key_name):
        self._key_binding_handlers[binding_name](key_state, key_name)

    def unregister_key_binding(self, keydef):
        binding_name = MPV._binding_name(keydef)
        self.command('disable-section', binding_name)
        self.command('define-section', binding_name, '')
        if callable(callback):
            del self._key_binding_handlers[binding_name]
            if not self._key_binding_handlers:
                self.unregister_message_handler('key-binding')

    # Convenience functions
    def play(self, filename):
        self.loadfile(filename)

    # Property accessors
    def _get_property(self, name, proptype=str, decode_str=False):
        fmt = {int:         MpvFormat.INT64,
               float:       MpvFormat.DOUBLE,
               bool:        MpvFormat.FLAG,
               str:         MpvFormat.STRING,
               bytes:       MpvFormat.STRING,
               commalist:   MpvFormat.STRING,
               MpvFormat.NODE: MpvFormat.NODE}[proptype]

        out = cast(create_string_buffer(sizeof(c_void_p)), c_void_p)
        outptr = byref(out)
        try:
            cval = _mpv_get_property(self.handle, name.encode('utf-8'), fmt, outptr)
            rv = MpvNode.node_cast_value(outptr, fmt, decode_str or proptype in (str, commalist))

            if proptype is commalist:
                rv = proptype(rv)

            if proptype is str:
                _mpv_free(out)
            elif proptype is MpvFormat.NODE:
                _mpv_free_node_contents(outptr)

            return rv
        except PropertyUnavailableError as ex:
            return None

    def _set_property(self, name, value, proptype=str):
        ename = name.encode('utf-8')
        if type(value) is bytes:
            _mpv_set_property_string(self.handle, ename, value)
        elif type(value) is bool:
            _mpv_set_property_string(self.handle, ename, b'yes' if value else b'no')
        elif proptype in (str, int, float):
            _mpv_set_property_string(self.handle, ename, str(proptype(value)).encode('utf-8'))
        else:
            raise TypeError('Cannot set {} property {} to value of type {}'.format(proptype, name, type(value)))

    # Dict-like option access
    def __getitem__(self, name, file_local=False):
        """ Get an option value """
        prefix = 'file-local-options/' if file_local else 'options/'
        return self._get_property(prefix+name)

    def __setitem__(self, name, value, file_local=False):
        """ Get an option value """
        prefix = 'file-local-options/' if file_local else 'options/'
        return self._set_property(prefix+name, value)

    def __iter__(self):
        return iter(self.options)

    def option_info(self, name):
        return self._get_property('option-info/'+name)

def commalist(propval=''):
    return str(propval).split(',')

node = MpvFormat.NODE

ALL_PROPERTIES = {
        'osd-level':                    (int,    'rw'),
        'osd-scale':                    (float,  'rw'),
        'loop':                         (str,    'rw'),
        'loop-file':                    (str,    'rw'),
        'speed':                        (float,  'rw'),
        'filename':                     (bytes,  'r'),
        'file-size':                    (int,    'r'),
        'path':                         (bytes,  'r'),
        'media-title':                  (bytes,  'r'),
        'stream-pos':                   (int,    'rw'),
        'stream-end':                   (int,    'r'),
        'length':                       (float,  'r'), # deprecated for ages now
        'duration':                     (float,  'r'),
        'avsync':                       (float,  'r'),
        'total-avsync-change':          (float,  'r'),
        'drop-frame-count':             (int,    'r'),
        'percent-pos':                  (float,  'rw'),
#        'ratio-pos':                    (float,  'rw'),
        'time-pos':                     (float,  'rw'),
        'time-start':                   (float,  'r'),
        'time-remaining':               (float,  'r'),
        'playtime-remaining':           (float,  'r'),
        'chapter':                      (int,    'rw'),
        'edition':                      (int,    'rw'),
        'disc-titles':                  (int,    'r'),
        'disc-title':                   (str,    'rw'),
#        'disc-menu-active':             (bool,   'r'),
        'chapters':                     (int,    'r'),
        'editions':                     (int,    'r'),
        'angle':                        (int,    'rw'),
        'pause':                        (bool,   'rw'),
        'core-idle':                    (bool,   'r'),
        'cache':                        (int,    'r'),
        'cache-size':                   (int,    'rw'),
        'cache-free':                   (int,    'r'),
        'cache-used':                   (int,    'r'),
        'cache-speed':                  (int,    'r'),
        'cache-idle':                   (bool,   'r'),
        'cache-buffering-state':        (int,    'r'),
        'paused-for-cache':             (bool,   'r'),
#        'pause-for-cache':              (bool,   'r'),
        'eof-reached':                  (bool,   'r'),
#        'pts-association-mode':         (str,    'rw'),
        'hr-seek':                      (str,    'rw'),
        'volume':                       (float,  'rw'),
        'volume-max':                   (int,    'rw'),
        'ao-volume':                    (float,  'rw'),
        'mute':                         (bool,   'rw'),
        'ao-mute':                      (bool,   'rw'),
        'audio-speed-correction':       (float,  'r'),
        'audio-delay':                  (float,  'rw'),
        'audio-format':                 (str,    'r'),
        'audio-codec':                  (str,    'r'),
        'audio-codec-name':             (str,    'r'),
        'audio-bitrate':                (float,  'r'),
        'packet-audio-bitrate':         (float,  'r'),
        'audio-samplerate':             (int,    'r'),
        'audio-channels':               (str,    'r'),
        'aid':                          (str,    'rw'),
        'audio':                        (str,    'rw'), # alias for aid
        'balance':                      (int,    'rw'),
        'fullscreen':                   (bool,   'rw'),
        'deinterlace':                  (str,    'rw'),
        'colormatrix':                  (str,    'rw'),
        'colormatrix-input-range':      (str,    'rw'),
#        'colormatrix-output-range':     (str,    'rw'),
        'colormatrix-primaries':        (str,    'rw'),
        'ontop':                        (bool,   'rw'),
        'border':                       (bool,   'rw'),
        'framedrop':                    (str,    'rw'),
        'gamma':                        (float,  'rw'),
        'brightness':                   (int,    'rw'),
        'contrast':                     (int,    'rw'),
        'saturation':                   (int,    'rw'),
        'hue':                          (int,    'rw'),
        'hwdec':                        (str,    'rw'),
        'panscan':                      (float,  'rw'),
        'video-format':                 (str,    'r'),
        'video-codec':                  (str,    'r'),
        'video-bitrate':                (float,  'r'),
        'packet-video-bitrate':         (float,  'r'),
        'width':                        (int,    'r'),
        'height':                       (int,    'r'),
        'dwidth':                       (int,    'r'),
        'dheight':                      (int,    'r'),
        'fps':                          (float,  'r'),
        'estimated-vf-fps':             (float,  'r'),
        'window-scale':                 (float,  'rw'),
        'video-aspect':                 (str,    'rw'),
        'osd-width':                    (int,    'r'),
        'osd-height':                   (int,    'r'),
        'osd-par':                      (float,  'r'),
        'vid':                          (str,    'rw'),
        'video':                        (str,    'rw'), # alias for vid
        'video-align-x':                (float,  'rw'),
        'video-align-y':                (float,  'rw'),
        'video-pan-x':                  (float,  'rw'),
        'video-pan-y':                  (float,  'rw'),
        'video-zoom':                   (float,  'rw'),
        'video-unscaled':               (bool,   'w'),
        'video-speed-correction':       (float,  'r'),
        'program':                      (int,    'w'),
        'sid':                          (str,    'rw'),
        'sub':                          (str,    'rw'), # alias for sid
        'secondary-sid':                (str,    'rw'),
        'sub-delay':                    (float,  'rw'),
        'sub-pos':                      (int,    'rw'),
        'sub-visibility':               (bool,   'rw'),
        'sub-forced-only':              (bool,   'rw'),
        'sub-scale':                    (float,  'rw'),
        'sub-bitrate':                  (float,  'r'),
        'packet-sub-bitrate':           (float,  'r'),
#        'ass-use-margins':              (bool,   'rw'),
        'ass-vsfilter-aspect-compat':   (bool,   'rw'),
        'ass-style-override':           (bool,   'rw'),
        'stream-capture':               (str,    'rw'),
        'tv-brightness':                (int,    'rw'),
        'tv-contrast':                  (int,    'rw'),
        'tv-saturation':                (int,    'rw'),
        'tv-hue':                       (int,    'rw'),
        'playlist-pos':                 (int,    'rw'),
        'playlist-pos-1':               (int,    'rw'), # ugh.
        'playlist-count':               (int,    'r'),
#        'quvi-format':                  (str,    'rw'),
        'seekable':                     (bool,   'r'),
        'seeking':                      (bool,   'r'),
        'partially-seekable':           (bool,   'r'),
        'playback-abort':               (bool,   'r'),
        'cursor-autohide':              (str,    'rw'),
        'audio-device':                 (str,    'rw'),
        'current-vo':                   (str,    'r'),
        'current-ao':                   (str,    'r'),
        'audio-out-detected-device':    (str,    'r'),
        'protocol-list':                (str,    'r'),
        'mpv-version':                  (str,    'r'),
        'mpv-configuration':            (str,    'r'),
        'ffmpeg-version':               (str,    'r'),
        'display-sync-active':          (bool,   'r'),
        'stream-open-filename':         (bytes,   'rw'), # Undocumented
        'file-format':                  (commalist,'r'), # Be careful with this one.
        'mistimed-frame-count':         (int,    'r'),
        'vsync-ratio':                  (float,  'r'),
        'vo-drop-frame-count':          (int,    'r'),
        'vo-delayed-frame-count':       (int,    'r'),
        'playback-time':                (float,  'rw'),
        'demuxer-cache-duration':       (float,  'r'),
        'demuxer-cache-time':           (float,  'r'),
        'demuxer-cache-idle':           (bool,   'r'),
        'idle':                         (bool,   'r'),
        'disc-title-list':              (commalist,'r'),
        'field-dominance':              (str,    'rw'),
        'taskbar-progress':             (bool,   'rw'),
        'on-all-workspaces':            (bool,   'rw'),
        'video-output-levels':          (str,    'r'),
        'vo-configured':                (bool,   'r'),
        'hwdec-current':                (str,    'r'),
        'hwdec-interop':                (str,    'r'),
        'estimated-frame-count':        (int,    'r'),
        'estimated-frame-number':       (int,    'r'),
        'sub-use-margins':              (bool,   'rw'),
        'ass-force-margins':            (bool,   'rw'),
        'video-rotate':                 (str,    'rw'),
        'video-stereo-mode':            (str,    'rw'),
        'ab-loop-a':                    (str,    'r'), # What a mess...
        'ab-loop-b':                    (str,    'r'),
        'dvb-channel':                  (str,    'w'),
        'dvb-channel-name':             (str,    'rw'),
        'window-minimized':             (bool,   'r'),
        'display-names':                (commalist, 'r'),
        'display-fps':                  (float,  'r'), # access apparently misdocumented in the manpage
        'estimated-display-fps':        (float,  'r'),
        'vsync-jitter':                 (float,  'r'),
        'video-params':                 (node,   'r', True),
        'video-out-params':             (node,   'r', True),
        'track-list':                   (node,   'r', False),
        'playlist':                     (node,   'r', False),
        'chapter-list':                 (node,   'r', False),
        'vo-performance':               (node,   'r', True),
        'filtered-metadata':            (node,   'r', False),
        'metadata':                     (node,   'r', False),
        'chapter-metadata':             (node,   'r', False),
        'vf-metadata':                  (node,   'r', False),
        'af-metadata':                  (node,   'r', False),
        'edition-list':                 (node,   'r', False),
        'disc-titles':                  (node,   'r', False),
        'audio-params':                 (node,   'r', True),
        'audio-out-params':             (node,   'r', True),
        'audio-device-list':            (node,   'r', True),
        'video-frame-info':             (node,   'r', True),
        'decoder-list':                 (node,   'r', True),
        'encoder-list':                 (node,   'r', True),
        'vf':                           (node,   'r', True),
        'af':                           (node,   'r', True),
        'options':                      (node,   'r', True),
        'file-local-options':           (node,   'r', True),
        'property-list':                (commalist,'r')}

def bindproperty(MPV, name, proptype, access, decode_str=False):
    getter = lambda self: self._get_property(name, proptype, decode_str)
    setter = lambda self, value: self._set_property(name, value, proptype)

    def barf(*args):
        raise NotImplementedError('Access denied')

    setattr(MPV, name.replace('-', '_'), property(getter if 'r' in access else barf, setter if 'w' in access else barf))

for name, (proptype, access, *args) in ALL_PROPERTIES.items():
    bindproperty(MPV, name, proptype, access, *args)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer UIEvent

Created on 2017-03-05
@author: Mellcap
'''

from mellplayer.ui import mell_ui, mell_lyric_ui

class UIEvent(object):

    # ===========================
    # Main UI
    # ===========================
    
    def handler_update_playInfo(self, play_info):
        mell_ui.update_play_info(play_info)

    def handler_update_title(self, items):
        mell_ui.update_title(items)

    def handler_show_error(self, err_msg):
        mell_ui.display_center(text=err_msg)

    # ===========================
    # Lyric UI
    # ===========================

    def handler_initial_lyric(self):
        mell_lyric_ui.initial_lyric()
        
    def handler_parse_lyric(self, origin_lyric):
        self.handler_initial_lyric()
        mell_lyric_ui.parse_lyric(origin_lyric=origin_lyric)

    def handler_roll_lyric(self, timestamp):
        mell_lyric_ui.roll(timestamp=timestamp)
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer UI

Created on 2017-02-21
@author: Mellcap
'''

import os
import re

from mellplayer.mell_logger import mell_logger


SONG_CATEGORIES = (
    '榜单', '流行', '摇滚', '民谣', '说唱', '轻音乐', '爵士', '乡村', '古典', '电子', '舞曲', '另类/独立',\
    '学习', '工作', '午休', '下午茶', '清晨', '夜晚',\
    '华语', '欧美', '日语', '韩语', '粤语', '小语种',\
    '怀旧', '清新', '浪漫', '性感', '伤感', '治愈', '放松', '孤独', '感动', '兴奋', '快乐', '安静', '思念', \
    '民族', '英伦', '金属', '朋克', '蓝调', '雷鬼', '世界音乐', '拉丁', 'New Age', '古风', '后摇', 'Bossa Nova',\
    '地铁', '驾车', '运动', '旅行', '散步', '酒吧',\
    '影视原声', 'ACG', '校园', '游戏', '70后', '80后', '90后', '00后',\
    '网络歌曲', 'KTV', '经典', '翻唱', '吉他', '钢琴', '器乐', '儿童',
)

# 所有颜色 https://upload.wikimedia.org/wikipedia/commons/1/15/Xterm_256color_chart.svg
FOREGROUND_COLOR = {       # 前景色
    'default'      : 249,
    'white'        : 15,
    'blue'         : 39,    
    'green'        : 118,
    'gray'         : 239,
    'red'          : 196,
    'pink'         : 201,
    'yellow'       : 220,
    'light_gray'   : 250,
}

BLANK_CONSTANT = 3
TERMINAL_SIZE = os.get_terminal_size()
BOLD_STR = '\033[1m'
MAX_LINES = len(SONG_CATEGORIES)
ALL_LINES = MAX_LINES + BLANK_CONSTANT

class UI(object):

    def __init__(self, ui_mode='home'):
        self.category_lines = SONG_CATEGORIES
        self.mark_index = 0
        self.play_index= 0
        self.play_info = ''
        
        self.top_index = 0
        self.screen_height = TERMINAL_SIZE.lines
        self.screen_width = TERMINAL_SIZE.columns
        self.base_title = self._get_base_title()
        self.title = self.base_title
        self.ui_mode = ui_mode

    # =====================
    # UI Displayer
    # =====================
    
    def _get_base_title(self):
        player_name = '%s%s' % (BOLD_STR, self.gen_color('MellPlayer', 'blue'))
        netease = self.gen_color('网易云音乐', 'red')
        divider = self.gen_color(data=r'\\')
        display_items = [player_name, netease]
        return (' %s ' % divider).join(display_items)

    def update_title(self, items=None):
        if not items:
            self.title = self.base_title
        else:
            divider = self.gen_color(data=r'\\')
            divider_play = self.gen_color(data=r'>>')
            items = [self.gen_color(data=i, color='pink') for i in items]
            extend_title = (' %s ' % divider).join(items)
            extend_title = ' %s %s' % (divider_play, extend_title)
            self.title = self.base_title + extend_title
        self.display()

    def display(self):
        '''
        UI 输出
        说明：多线程终端输出有问题，在每行结尾加\r
        '''
        display_lines = ['\r']
        display_title = '\n%s%s' % (' '*5, self.title)
        display_lines.append(display_title)
        top_index = self.top_index
        bottom_index = (self.screen_height - BLANK_CONSTANT) + top_index

        for index, category in enumerate(self.category_lines[top_index: bottom_index]):
            # mark_index            
            is_markline = True if (index + self.top_index) == self.mark_index else False
            category = self.gen_category(category, is_markline)
            # play_index
            play_info = ''
            is_playline = True if (index + self.top_index) == self.play_index else False
            if is_playline:
                play_info = self.gen_playline()

            complete_line = '%s%s%s' % (category, ' '*10, play_info)
            display_lines.append(complete_line)

        if ALL_LINES < self.screen_height:
            # fill_blanks
            display_lines = self.fill_blanks(display_lines)
        # add tail
        display_lines = self.add_tail(source_list=display_lines, tail='\r')
        print('\n'.join(display_lines))
        
    def gen_category(self, category, is_markline=False):
        '''
        生成分类样式
        '''
        if is_markline:
            category = self.gen_mark(category)
            category = self.gen_color(data=category, color='yellow')
        else:
            category = '%s%s' % (' '*5, category)
            category = self.gen_color(data=category, color='')
        return category

    def gen_mark(self, category):
        return '  ➣  %s' % category

    def gen_playline(self):
        '''
        生成歌曲展示行
        '''
        complete_info = [self.gen_color(data=p, color='yellow') for p in self.play_info]
        divider = self.gen_color(data='|', color='')
        return ('  %s  ' % divider).join(complete_info)

    def display_center(self, text):
        display_lines = ['\r']
        display_title = '\n%s%s' % (' '*5, self.title)
        display_lines.append(display_title)
        blank_length = self.screen_height - BLANK_CONSTANT
        blank_lines = [' ' for i in range(blank_length)]
        blank_lines[int(blank_length / 2)] = str_center(string=text, screen_width=self.screen_width)
        display_lines.extend(blank_lines)
        display_lines = self.add_tail(source_list=display_lines, tail='\r')
        print('\n'.join(display_lines) + '\r')

    # =====================
    # UI Controller
    # =====================
    
    def next_line(self):
        '''
        下一行
        '''
        if self.mark_index < (MAX_LINES - 1):
            self.mark_index += 1
            bottom_index = (self.screen_height - BLANK_CONSTANT) + self.top_index
            if self.mark_index > (bottom_index - 1):
                self.top_index += 1
        self.display()

    def prev_line(self):
        '''
        上一行
        '''
        if self.mark_index > 0:
            self.mark_index -= 1
            if self.mark_index < self.top_index:
                self.top_index -= 1
        self.display()

    def update_play_index(self):
        '''
        更新歌曲信息展示在光标选定行
        '''
        self.play_index = self.mark_index

    def update_play_info(self, play_info):
        '''
        更新歌曲信息
        '''
        if type(play_info) is str:
            play_info = [play_info]
        self.play_info = play_info
        self.display()

    # =====================
    # Utils
    # =====================
    
    def gen_color(self, data, color='default'):
        '''
        参考地址:http://blog.csdn.net/gatieme/article/details/45439671
        但是目前用不到这么多类型，目前只用前景色
        '''
        color_code = FOREGROUND_COLOR.get(color, 246)
        data = "\001\033[38;5;%sm\002%s\001\033[0m\002" % (color_code, data)
        return data

    def add_tail(self, source_list, tail):
        return map(lambda x: '%s%s' % (str(x), tail), source_list)

    def fill_blanks(self, display_lines, all_lines=ALL_LINES, position='after'):
        '''
        补全空白行
        '''
        delta_lines = self.screen_height - all_lines
        delta_blanks = [' ' for i in range(delta_lines)]
        if position == 'after':
            display_lines += delta_blanks
        elif position == 'before':
            display_lines = delta_blanks + display_lines
        return display_lines

    
# =====================
# HelpUI
# =====================

HELP_LINES = {
    'help_space_1':    '',
    'control_move':    '操作',
    'next_line':       '[j]     [Next Line]         --->  下',
    'prev_line':       '[k]     [Prev Line]         --->  上',
    'quit':            '[q]     [Quit]              --->  退出',
    'help_space_2':    '',
    'control_music':   '音乐',
    'space':           '[space] [Start/Pause]       --->  播放／暂停',
    'next_song':       '[n]     [Next Song]         --->  下一曲',
    'prev_song':       '[p]     [Prev Song]         --->  上一曲',
    'next_playlist':   '[f]     [Forward Playlist]  --->  下个歌单',
    'prev_playlist':   '[b]     [Backward Playlist] --->  上个歌单',
    'help_space_3':    '',
    'control_volume':  '音量',
    'reduce_volume':   '[-]     [Reduce Volume]     --->  减小音量',
    'increase_volume': '[=]     [Increase Volume]   --->  增加音量',
    'mute':            '[m]     [Mute]              --->  静音',
    'help_space_4': '',
    'control_lyric': '歌词',
    'lyric':           '[l]     [Show/Hide Lyric]   --->  显示／关闭歌词',
    'help_space_5': '',
    'control_help': '帮助',
    'help':            '[h]     [Show/Hide Help]    --->  显示／关闭帮助'
}

class HelpUI(UI):
    
    def __init__(self):
        super(HelpUI, self).__init__(ui_mode='help')

    def display(self):
        display_lines = ['\r']
        display_title = '\n%s%s' % (' '*5, self.title)
        display_lines.append(display_title)
        for key, help_line in HELP_LINES.items():
            colored_help_line = self.color_line(key=key, line=help_line)
            display_lines.append('%s%s' % (' '*5, colored_help_line))

        # fill blanks
        all_lines = len(HELP_LINES) + BLANK_CONSTANT
        if all_lines < self.screen_height:
            display_lines = self.fill_blanks(display_lines, all_lines=all_lines)
        # add tail
        display_lines = self.add_tail(source_list=display_lines, tail='\r')
        print('\n'.join(display_lines))

    def color_line(self, key, line):
        if key.startswith('control'):
            return self.gen_color(data='%s%s' % (BOLD_STR, line), color='yellow')
        else:
            return self.gen_color(data=line, color='light_gray')

# =====================
# LyricUI
# =====================

class LyricUI(UI):
    
    def __init__(self):
        super(LyricUI, self).__init__(ui_mode='lyric')
        self.has_lyric = True
        self.lyric_times = None
        self.lyric_lines = ''
        self.lyric_display_lines = ''

    def parse_lyric(self, origin_lyric):
        '''
        解析歌词
        '''
        if origin_lyric == 'no_lyric':
            self.has_lyric = False
        else:
            compiler = re.compile('\[(.+)\](.+?)\n')
            format_lyric = compiler.findall(origin_lyric)
            if format_lyric:
                self.lyric_times = [format_minute2second(l[0]) for l in format_lyric]
                self.lyric_lines = [l[1] for l in format_lyric]
            else:
                self.has_lyric = False

    def display(self):
        display_lines = ['\r']
        display_title = '\n%s%s' % (' '*5, self.title)
        display_lines.append(display_title)
        if not self.has_lyric:
            text = '求歌词...'
            self.display_center(text=text, display_lines=display_lines)
        elif not self.lyric_display_lines:
            text = '貌似Ta还没开嗓...'
            self.display_center(text=text, display_lines=display_lines)
        elif self.lyric_lines:
            self.display_lyric(display_lines=display_lines)

    def make_display_lines(self, display_lines):
        '''
        生成展示歌词
        '''
        all_lines = len(self.lyric_display_lines) + BLANK_CONSTANT
        lyric_display_lines = self.lyric_display_lines
        if all_lines >= self.screen_height:
            display_index = all_lines - self.screen_height
            lyric_display_lines = lyric_display_lines[display_index:]
        else:
            lyric_display_lines = self.fill_blanks(lyric_display_lines, all_lines=all_lines, position='before')
        return lyric_display_lines

    def display_lyric(self, display_lines):
        lyric_display_lines = self.make_display_lines(display_lines)
        for line in lyric_display_lines:
            # 居中
            line = str_center(string=line, screen_width=self.screen_width)
            line = self.gen_color(data=line, color='light_gray')
            display_lines.append(line)
        display_lines = self.add_tail(source_list=display_lines, tail='\r')
        print('\n'.join(display_lines) + '\r')

    def display_center(self, text, display_lines):
        blank_length = self.screen_height - BLANK_CONSTANT
        blank_lines = [' ' for i in range(blank_length)]
        blank_lines[int(blank_length / 2)] = str_center(string=text, screen_width=self.screen_width)
        display_lines.extend(blank_lines)
        display_lines = self.add_tail(source_list=display_lines, tail='\r')
        print('\n'.join(display_lines) + '\r')

    def roll(self, timestamp):
        lyric_times = self.lyric_times
        if lyric_times:
            if timestamp in lyric_times:
                lyric_index = lyric_times.index(timestamp) + 1
            else:
                lyric_times_copy = lyric_times[:]
                lyric_times_copy.append(timestamp)
                lyric_times_copy.sort()
                lyric_index = lyric_times_copy.index(timestamp)
            self.lyric_display_lines = self.lyric_lines[:lyric_index]
            self.display()

    def initial_lyric(self):
        self.has_lyric = True
        self.lyric_times = None
        self.lyric_lines = ''
        self.lyric_display_lines = ''

        
# =====================
# Basic Method
# =====================

def format_minute2second(timestamp):
    stamp_list = timestamp[:5].split(':')
    return int(stamp_list[0]) * 60 + int(stamp_list[1])

def str_len(string):
    '''
    计算string实际占字符长
    '''
    row_l = len(string)
    utf8_l = len(string.encode('utf-8'))
    return int((utf8_l - row_l) / 2 + row_l)

def str_center(string, screen_width, fill_char=None):
    '''
    字符居中
    '''
    delta_width = screen_width - (str_len(string) - len(string))
    if fill_char:
        return string.center(delta_width, fill_char)
    return string.center(delta_width)


# =====================
# Instance
# =====================

mell_ui = UI()
mell_help_ui = HelpUI()
mell_lyric_ui = LyricUI()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer Starter

Created on 2017-03-05
@author: Mellcap
'''

from mellplayer.controller import mell_ui, initial_player
from mellplayer.watcher import time_watcher, key_watcher

def main():
    print('Initial Player...')
    initial_player()
    time_watcher()
    key_watcher()
    mell_ui.display()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer Controller

Created on 2017-03-05
@author: Mellcap
'''
import threading

from mellplayer.player import mell_player
from mellplayer.ui import mell_ui, mell_help_ui, mell_lyric_ui, SONG_CATEGORIES
from mellplayer.mell_logger import mell_logger


# ===========================
# Controller Handler
# ===========================

def handler_space():
    current_category = SONG_CATEGORIES[mell_ui.mark_index]
    if mell_player.category == current_category:
        handler_play()
    else:
        # change UI play_index 
        mell_ui.update_play_index()
        # change playlist category
        mell_player.switch_category(new_category=current_category)


def handler_next_line():
    mell_ui.next_line()

def handler_prev_line():
    mell_ui.prev_line()

def handler_play():
    mell_player.start_or_pause()

def handler_next_song():
    mell_player.next_song()

def handler_prev_song():
    mell_player.prev_song()

def handler_next_playlist():
    mell_player.next_playlist()

def handler_prev_playlist():
    mell_player.prev_playlist()

def handler_reduce_volume():
    mell_player.reduce_volume()

def handler_increase_volume():
    mell_player.increase_volume()

def handler_mute_volume():
    mell_player.mute_volume()

def handler_lyric():
    if mell_ui.ui_mode == 'home' and mell_player.time_remaining:
        mell_ui.ui_mode = 'lyric'
        handler_lyric_display()
    elif mell_ui.ui_mode == 'lyric':
        mell_ui.display()
        mell_ui.ui_mode = 'home'

def handler_lyric_display():
    song_id = mell_player.playlist_ids[mell_player.playlist_index]
    if mell_player.lyric_id != song_id:
        mell_lyric_ui.initial_lyric()
        mell_player.get_lyric_detail()
    mell_lyric_ui.display()

def handler_help():
    if mell_ui.ui_mode == 'home':
        mell_help_ui.display()
        mell_ui.ui_mode = 'help'
    elif mell_ui.ui_mode == 'help':
        mell_ui.display()
        mell_ui.ui_mode = 'home'

def handler_quit():
    mell_player.terminate()


# ===========================
# Initial Player
# ===========================

def i_player():
    current_category = SONG_CATEGORIES[mell_ui.mark_index]
    mell_player.switch_category(new_category=current_category)
    # handler_update_playInfo()

def initial_player():
    initPlayer_thread = threading.Thread(target=i_player)
    initPlayer_thread.start()




#!/usr/bin/env python

"""
Copyright (c) 2006-2015 sqlmap developers (http://sqlmap.org/)
See the file 'doc/COPYING' for copying permission
"""


class _Getch(object):
    """
    Gets a single character from standard input.  Does not echo to
    the screen (reference: http://code.activestate.com/recipes/134892/)
    """
    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            try:
                self.impl = _GetchMacCarbon()
            except(AttributeError, ImportError):
                self.impl = _GetchUnix()

    def __call__(self):
        return self.impl()


class _GetchUnix(object):
    def __init__(self):
        import tty

    def __call__(self):
        import sys
        import termios
        import tty

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows(object):
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch()


class _GetchMacCarbon(object):
    """
    A function which returns the current ASCII key that is down;
    if no ASCII key is down, the null string is returned.  The
    page http://www.mactech.com/macintosh-c/chap02-1.html was
    very helpful in figuring out how to do this.
    """
    def __init__(self):
        import Carbon
        Carbon.Evt  # see if it has this (in Unix, it doesn't)

    def __call__(self):
        import Carbon
        if Carbon.Evt.EventAvail(0x0008)[0] == 0:  # 0x0008 is the keyDownMask
            return ''
        else:
            #
            # The event contains the following info:
            # (what,msg,when,where,mod)=Carbon.Evt.GetNextEvent(0x0008)[1]
            #
            # The message (msg) contains the ASCII char which is
            # extracted with the 0x000000FF charCodeMask; this
            # number is converted to an ASCII character with chr() and
            # returned
            #
            (what, msg, when, where, mod) = Carbon.Evt.GetNextEvent(0x0008)[1]
            return chr(msg & 0x000000FF)


getch = _Getch()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer Watcher

Created on 2017-03-05
@author: Mellcap
'''

import sys
import time
import queue
import threading

from mellplayer.controller import *
from mellplayer.utils import getch

KEY_QUEUE = queue.Queue()
CONFIG = {
    # 主页
    'q': 'quit',
    'j': 'next_line',
    'k': 'prev_line',
    # 音乐
    ' ': 'space',
    'n': 'next_song',
    'p': 'prev_song',
    'f': 'next_playlist',
    'b': 'prev_playlist',
    # 音量
    '-': 'reduce_volume',
    '=': 'increase_volume',
    'm': 'mute_volume',
    # 歌词
    'l': 'lyric',
    # 帮助
    'h': 'help'
}

# ===========================
# Key Watcher
# ===========================

def k_watcher():
    while 1:
        key = getch.getch()
        KEY_QUEUE.put(key)
        if key == 'q':
            mell_player.is_quit = True
            break

def k_executor():
    while 1:
        key = KEY_QUEUE.get()
        action = CONFIG.get(key, None)
        if action:
            func = 'handler_%s' % action
            eval(func)()
            if action == 'quit':
                break

def key_watcher():
    k_watcher_thread = threading.Thread(target=k_watcher)
    k_executor_thread = threading.Thread(target=k_executor)

    k_watcher_thread.start()
    k_executor_thread.start()
    k_watcher_thread.join()
    k_executor_thread.join()


# ===========================
# Time Watcher
# ===========================

def t_watcher():
    while not mell_player.is_quit:
        if not mell_player.pause:
            time_pos = mell_player.time_pos
            time_remain = mell_player.time_remaining
            if time_pos and time_remain:
                if mell_ui.ui_mode == 'lyric':
                    mell_lyric_ui.roll(int(time_pos))
                if time_remain <= 2:
                    handler_next_song()
                timestamp = format_timestamp(time_pos, time_remain)
                show_footer(timestamp=timestamp)
        time.sleep(1)

def format_timestamp(time_pos, time_remain):
    total_time = time_pos + time_remain
    format_time_pos = '%02d:%02d' % divmod(time_pos, 60)
    format_time_total = '%02d:%02d' % divmod(total_time, 60)
    return ' / '.join((format_time_pos, format_time_total))

def show_footer(timestamp):
    timestamp = mell_ui.gen_color(data=timestamp, color='blue')
    footer = timestamp.rjust(mell_ui.screen_width + 13) + '\r'
    sys.stdout.write(footer)
    sys.stdout.flush()
            

def time_watcher():
    t_watcher_thread = threading.Thread(target=t_watcher)
    t_watcher_thread.start()




import base64
import json
import os
import binascii

from Crypto.Cipher import AES

modulus = '00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7b725' \
          '152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280104e0312ecbda' \
          '92557c93870114af6c9d05c4f7f0c3685b7a46bee255932575cce10b424d813cfe4' \
          '875d3e82047b97ddef52741d546b8e289dc6935b3ece0462db0a22b8e7'
nonce = '0CoJUm6Qyw8W8jud'
pubKey = '010001'


def encrypted_request(text):
    text = json.dumps(text)
    secKey = createSecretKey(16)
    encText = aesEncrypt(aesEncrypt(text, nonce), secKey)
    encSecKey = rsaEncrypt(secKey, pubKey, modulus)
    data = {'params': encText, 'encSecKey': encSecKey}
    return data


def aesEncrypt(text, secKey):
    pad = 16 - len(text) % 16
    text = text + chr(pad) * pad
    encryptor = AES.new(secKey, 2, '0102030405060708')
    ciphertext = encryptor.encrypt(text)
    ciphertext = base64.b64encode(ciphertext).decode('utf-8')
    return ciphertext


def rsaEncrypt(text, pubKey, modulus):
    text = text[::-1]
    rs = pow(int(binascii.hexlify(text), 16), int(pubKey, 16), int(modulus, 16))
    return format(rs, 'x').zfill(256)


def createSecretKey(size):
    return binascii.hexlify(os.urandom(size))[:16]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer Player Deco

Created on 2017-03-08
@author: Mellcap
'''

# ===========================
# Deco
# ===========================

def show_changing_text(func):
    '''
    加载歌曲显示
    '''
    def wrapper(*args, **kw):
        # args[0] == player
        p = args[0]
        p.show_song_changing()
        return func(*args, **kw)
    return wrapper

def show_song_info_text(func):
    '''
    歌曲详情显示
    '''
    def wrapper(*args, **kw):
        func(*args, **kw)
        p = args[0]
        while 1:
            if p.time_pos:
                p.show_song_info()
                break
            time.sleep(1)
    return wrapper

def update_title_text(func):
    '''
    更新title
    '''
    def wrapper(*args, **kw):
        # args[0] == player
        func(*args, **kw)
        p = args[0]
        p.update_title()
    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer logger

Created on 2017-03-06
@author: Mellcap
'''

import os
import logging

from mellplayer.directory import BASE_DIRECTORY

LOG_FILE = os.path.join(BASE_DIRECTORY, 'mell_logger.log')

 
# create logger
mell_logger = logging.getLogger('mell_logger')  
mell_logger.setLevel(logging.DEBUG)
 
# define handler write in file
fh = logging.FileHandler(LOG_FILE)
fh.setLevel(logging.DEBUG)
 
# define formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s \n- %(message)s')
fh.setFormatter(formatter)
 
# add handler
mell_logger.addHandler(fh)
 




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer Player

Created on 2017-02-20
@author: Mellcap
'''

import os
import time
import threading
import datetime

from mellplayer.utils.mpv import MPV
from mellplayer.api import Netease
from mellplayer.directory import BASE_DIRECTORY
from mellplayer.event.ui_event import UIEvent
from mellplayer.deco import show_changing_text, show_song_info_text, update_title_text
from mellplayer.mell_logger import mell_logger

PLAYLIST_MAX = 50
DEFAULT_VOLUME = 100
# PLAYLIST_FILE = os.path.join(BASE_DIRECTORY, 'playlist.m3u')
NeteaseApi = Netease()
UiEvent = UIEvent()


class Player(MPV):

    def __init__(self, *extra_mpv_flags, log_handler=None, start_event_thread=True, **extra_mpv_opts):
        super(Player, self).__init__(*extra_mpv_flags, log_handler=log_handler, start_event_thread=start_event_thread, **extra_mpv_opts)
        self.category = None
        self.category_playlist_ids = None
        self.category_playlist_index = 0
        self.playlist_ids = None
        self.playlist_index = 0
        self.playlist_detail = None
        self.song_info = ''
        self.song_br = 0
        self.lyric_id = 0
        self.is_quit = False

    # ===========================
    # Player Controller
    # ===========================

    def start_or_pause(self):
        self.pause = not self.pause

    def switch_song(self, action='next'):
        '''
        action: next/prev
        '''
        if action == 'next':
            self.playlist_next()
        elif action == 'prev':
            self.playlist_prev()

    @show_changing_text
    def next_song(self):
        # self.playlist_next()
        playlist_ids = self.playlist_ids
        if playlist_ids:
            self.playlist_index += 1
            if self.playlist_index >= len(playlist_ids):
                self.playlist_index = 0
            self.run_player()

    @show_changing_text
    def prev_song(self):
        # self.playlist_prev()
        playlist_ids = self.playlist_ids
        if playlist_ids:
            self.playlist_index -= 1
            if self.playlist_index < 0:
                self.playlist_index = len(self.playlist_ids) - 1
            self.run_player()

    def switch_playlist(self):
        '''
        action: next/prev
        '''
        if action == 'next':
            self.next_playlist()
        elif action == 'prev':
            self.prev_playlist()

    @show_changing_text
    def next_playlist(self):
        category_playlist_ids = self.category_playlist_ids
        if category_playlist_ids:
            self.category_playlist_index += 1
            if self.category_playlist_index >= len(self.category_playlist_ids):
                self.category_playlist_index = 0
            self.run_playlist()

    @show_changing_text
    def prev_playlist(self):
        category_playlist_ids = self.category_playlist_ids
        if category_playlist_ids:
            self.category_playlist_index -= 1
            if self.category_playlist_index < 0:
                self.category_playlist_index = len(self.category_playlist_ids) - 1
            self.run_playlist()

    @show_changing_text
    def switch_category(self, new_category):
        self.category = new_category
        self.get_category_playlist_ids()
        self.run_playlist()
    
    # ===========================
    # Play Info
    # ===========================
    
    def get_category_playlist_ids(self):
        '''
        获取该类别-歌单id列表
        '''
        category = self.category or '全部'
        # initial category_playlist_index
        self.category_playlist_index = 0
        # get data
        data = NeteaseApi.category_playlists(category=category)
        category_playlist_ids = NeteaseApi.parse_info(data=data, parse_type='category_playlists')
        self.category_playlist_ids = tuple(category_playlist_ids)
    
    def get_playlist(self):
        '''
        获取该歌单-歌曲id列表
        '''
        playlist_id = self.category_playlist_ids[self.category_playlist_index]
        data = NeteaseApi.playlist_detail(playlist_id)
        playlist_ids, playlist_detail = NeteaseApi.parse_info(data=data, parse_type='playlist_detail')
        self.playlist_ids = playlist_ids
        self.playlist_detail = playlist_detail
        self.update_playlist_url()

    def update_playlist_url(self):
        '''
        旧的mp3_url有很多无法播放，采用新的接口
        新接口mp3_url：有时间戳，经过一段时间失效，已在logger中refresh_playlist
        '''
        song_ids = self.playlist_ids
        data = NeteaseApi.song_detail_new(song_ids)
        song_details = NeteaseApi.parse_info(data=data, parse_type='song_detail_new')
        song_urls = list(map(lambda x: x['song_url'], song_details.values()))
        if not any(song_urls):
            err_msg = 'Ooops! The song API only supports the Chinese Mainland visit...'
            UiEvent.handler_show_error(err_msg)
            self.terminate()
        else:
            for song_id in self.playlist_detail:
                song_detail = song_details.get(song_id, None)
                if not song_detail:
                    continue
                song_info = {
                    'song_url': song_detail.get('song_url', None),
                    'song_br': song_detail.get('song_br', None)
                }
                self.playlist_detail[song_id].update(song_info)

    def get_lyric_detail(self):
        '''
        获取歌词详情
        '''
        song_id = self.playlist_ids[self.playlist_index]
        if self.lyric_id != song_id:
            self.lyric_id = song_id
            data = NeteaseApi.lyric_detail(song_id=song_id)
            lyric_detail = NeteaseApi.parse_info(data=data, parse_type='lyric_detail')['lyric']
            UiEvent.handler_parse_lyric(origin_lyric=lyric_detail)

    def update_song_info(self):
        '''
        更新歌曲信息：歌名 & 歌手名
        '''
        if self.playlist_ids and self.playlist_detail:
            play_detail = self.playlist_detail.get(self.playlist_ids[self.playlist_index], None)
            if play_detail:
                self.song_info = [play_detail.get('song_name', ''), play_detail.get('song_artists', '')]
                self.song_br = play_detail.get('song_br', 0)

    def get_play_info(self):
        '''
        待作废
        '''
        self.update_song_info()
        return self.song_info

    def show_song_info(self):
        '''
        歌曲信息、音量等信息
        '''
        self.update_song_info()
        UiEvent.handler_update_playInfo(self.song_info)
        self.update_title()

    def update_title(self):
        '''
        更新title: 码率和音量
        '''
        song_br = '%s%s' % (int(int(self.song_br)/1000), 'Kbps')
        volume = 'Volume: %s%s' % (int(self.get_volume()), '%')
        UiEvent.handler_update_title(items=[song_br, volume])

    def show_song_changing(self):
        '''
        加载歌曲loading
        '''
        changing_text = '加载歌曲中...'
        UiEvent.handler_update_playInfo(changing_text)

    def run_playlist(self):
        '''
        启动播放列表
        '''
        # initial playlist_index
        self.playlist_index = 0
        self.get_playlist()
        self.run_player()

    # @show_song_info_text
    def run_player(self):
        '''
        启动播放器
        '''
        if self.playlist_detail and self.playlist_ids:
            song_id = self.playlist_ids[self.playlist_index]
            song_info = self.playlist_detail.get(song_id, {})
            if not song_info:
                mell_logger.error('Can not get song_info, song_id: %s' % song_id)
                pass
            song_url = song_info.get('song_url', None)
            if not song_url:
                mell_logger.error('Can not get song_url, song_id: %s' % song_id)
                self.next_song()
            else:
                self.play(song_url)
                self.show_song_info()


    # ===========================
    # Volume Controller
    # ===========================

    @update_title_text
    def reduce_volume(self, step=10):
        '''
        减小音量
        '''
        volume = max(self.volume - step, 0)
        self.volume = volume

    @update_title_text
    def increase_volume(self, step=10):
        '''
        增加音量
        '''
        volume = min(self.volume + step, 100)
        self.volume = volume

    @update_title_text
    def mute_volume(self):
        '''
        静音
        '''
        self.mute = not self.mute

    def get_volume(self):
        return self.volume or DEFAULT_VOLUME


    # ===========================
    # Playlist 
    # ===========================
    # ***Have some bugs
    
    def init_playlist(self):
        '''
        playlist会发生闪退问题，暂时自己控制播放列表
        '''
        if os.path.exists(PLAYLIST_FILE):
            self.loadlist(PLAYLIST_FILE)

    def loop_playlist(self):
        '''
        循环播放
        '''
        self._set_property('loop', True)
        
    def save_playlist(self):
        '''
        保存播放列表为m3u格式
        '''
        playlist = []
        m3u_title = '#EXTM3U\n'
        playlist.append(m3u_title)
        for song in self.playlist_detail:
            song_detail = '#EXTINF:,%s\n%s\n' % (song['song_id'], song['song_url'])
            playlist.append(song_detail)
        with open(PLAYLIST_FILE, 'w') as f:
            for line in playlist:
                f.write(line)
        return True
    


# ===========================
# Instance
# ===========================

def player_logger(loglevel, component, message):
    if loglevel == 'error':
        # print('[{}] {}: {}\r'.format(loglevel, component, message))
        refresh_playlist()
    # mell_logger.info('[%s] %s: %s' % (loglevel, component, message))
        
mell_player = Player(log_handler=player_logger, ytdl=True)

def refresh_playlist():
    mell_player.get_playlist()
    mell_player.run_player()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer Setup

Created on 2017-03-09
@author: Mellcap
'''

from setuptools import setup, find_packages
VERSION = '0.1.1.4'

setup(
    name='MellPlayer',
    version=VERSION,
    packages=find_packages(),
    install_requires=[
        'requests',
        'pycrypto'
    ],

    entry_points={
        'console_scripts': [
            'mellplayer = mellplayer.start:main'
        ],
    },

    license='MIT',
    author='Mellcap',
    author_email='imellcap@gmail.com',
    url='https://github.com/Mellcap/MellPlayer',
    description='A tiny terminal player based on Python',
    keywords=['mellplayer', 'terminal', 'playlist', 'music', 'cli', 'player'],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Netease Music API

Created on 2017-02-19
@author: Mellcap
'''
import requests
import json

from mellplayer.utils.encrypt_utils import encrypted_request
from mellplayer.mell_logger import mell_logger


class Netease(object):

    def __init__(self):
        self.playlist_categories = []

    def _request(self, url, method='GET', is_raw=True, data=None):
        '''
        对requests简单封装
        '''
        headers = {'appver': '2.0.2', 'Referer': 'http://music.163.com'}
        if method == 'GET':
            result = requests.get(url=url, headers=headers)
        elif method == 'POST' and data:
            result = requests.post(url=url, data=data, headers=headers)
        # if request failed, return False
        if not result.ok:
            return False
        result.encoding = 'UTF-8'
        if is_raw:
            return json.loads(result.text)
        return result.text

    def playlist_categories(self):
        '''
        分类歌单
        http://music.163.com/discover/playlist/
        '''
        url = 'http://music.163.com/discover/playlist/'
        result = self._request(url)
        return result

    def category_playlists(self, category='流行', offset=0, limit=50, order='hot', total='false'):
        '''
        分类详情
        http://music.163.com/api/playlist/list?cat=流行&order=hot&offset=0&total=false&limit=50
        '''
        url = 'http://music.163.com/api/playlist/list?cat=%s&order=%s&offset=%s&total=%s&limit=%s' % (category, order, offset, total, limit)
        result = self._request(url)
        return result

    def playlist_detail(self, playlist_id):
        '''
        歌单详情
        http://music.163.com/api/playlist/detail?id=xxx
        '''
        url = 'http://music.163.com/api/playlist/detail?id=%s' % playlist_id
        result = self._request(url)
        return result

    
    def song_detail(self, song_ids):
        '''
        歌曲详情
        http://music.163.com/api/song/detail?ids=[xxx, xxx]
        '''
        url = 'http://music.163.com/api/song/detail?ids=%s' % song_ids
        result = self._request(url)
        return result

    def song_detail_new(self, song_ids):
        url = 'http://music.163.com/weapi/song/enhance/player/url?csrf_token='
        data = {'ids': song_ids, 'br': 320000, 'csrf_token': ''}
        data = encrypted_request(data)
        result = self._request(url, method="POST", data=data)
        return result

   
    def lyric_detail(self, song_id):
        '''
        歌词详情
        http://music.163.com/api/song/lyric?os=osx&id=xxx&lv=-1&kv=-1&tv=-1
        '''
        url = 'http://music.163.com/api/song/lyric?os=osx&id=%s&lv=-1&kv=-1&tv=-1' % song_id
        result = self._request(url)
        return result


    def parse_info(self, data, parse_type):
        '''
        解析信息
        '''
        res = None
        if parse_type == 'category_playlists':
            res = [d['id'] for d in data['playlists']]
        elif parse_type == 'playlist_detail':
            tracks = data['result']['tracks']
            playlist_ids = [t['id'] for t in tracks]
            playlist_detail = {t['id']: {
                'song_id': t['id'],
                'song_name': t['name'],
                'song_url': t['mp3Url'],
                'song_artists': ' & '.join(map(lambda a: a['name'], t['artists']))
            } for t in tracks}
            res = (playlist_ids, playlist_detail)
        elif parse_type == 'lyric_detail':
            if 'lrc' in data:
                res = {
                    'lyric': data['lrc']['lyric']
                }
            else:
                res = {
                    'lyric': 'no_lyric'
                }
        elif parse_type == 'song_detail_new':
            res = {d['id']: {
                'song_url': d['url'],
                'song_br': d['br']
            } for d in data['data']}
        return res
                
            



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
MellPlayer Directory

Created on 2017-02-23
@author: Mellcap
'''

import os

BASE_DIRECTORY = os.path.join(os.path.expanduser('~'), '.MellPlayer')

class Directory(object):

    def create_directory(self, directory=None):
        if not directory:
            directory = BASE_DIRECTORY
        if not os.path.exists(directory):
            os.makedirs(directory)

# ===========================
# Instance
# ===========================

mell_directory = Directory()
mell_directory.create_directory()
